package com.perforce.api;

import java.io.*;
import java.util.*;

/*
 * Copyright (c) 2001, Perforce Software, All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/**
 * Class that contains static utility methods.
 *
 * @see HashDecay
 * @author <a href="mailto:david@markley.cc">David Markley</a>
 * @version $Date: 2002/05/15 $ $Revision: #1 $
 */
public final class Utils {

  /**
   * Initializes the package, in order to avoid some arbitrary JVM problems
   * that have been encountered. This is a hack and should not have to be done
   * if certain overbearing empires starting with the letter M built a fully
   * compliant JVM.
   */
  public static void initPackage() {
    (new Branch()).getCache();
    (new Change()).getCache();
    (new Client()).getCache();
    (new DirEntry()).getCache();
    (new FileEntry()).getCache();
    (new Job()).getCache();
    (new Label()).getCache();
    (new User()).getCache();
    Properties props = System.getProperties();
  }

  /**
   * Check to see if the current Java Virtual Machine is made by Microsoft
   *
   * @return boolean true if the java.vendor property comtains the word 'Microsoft'
   */
  public static boolean isMSJVM() {
    return (System.getProperty("java.vendor").indexOf("Microsoft") != -1);
  }
  /**
   * Check to see if the current operating system is a Windows OS
   *
   * @return boolean true if the os.name property comtains the word 'Windows'
   */
  public static boolean isWindows() {
    return (System.getProperty("os.name").indexOf("Windows") != -1);
  }
 
  /**
   * Returns true if the <code>path</code> matches the <code>wildpath</code>.
   * Only perforce wildcards are considered in the <code>wildpath</code>.
   */
  public static boolean wildPathMatch(String wildpath, String path) {
    wildpath = wildpath.trim();
    path = path.trim();
    boolean match = true;
    int i = 0, j = 0, plen = path.length();
    char wc, pc;
    for (i = 0; i < wildpath.length(); i++) {
      if (j >= plen) {
	match = false;
	break;
      }
      if ('%' == (wc = wildpath.charAt(i))) {
	wc = wildpath.charAt(++i);
	if ('0' > wc || '9' < wc) {
	  match = false;
	  break;
	}
	while ('/' != path.charAt(j) && j < plen) { j++; }
	continue;
      } else if ('*' == wc) {
	while ('/' != path.charAt(j) && j < plen) { j++; }
	continue;
      }
      if ('.' == wc && wildpath.regionMatches(i, "...", 0, 3)) {
	i += 3;

      }
      if (path.charAt(j++) != wc) {
	match = false;
	break;
      }
    }
    return match;
  }

  /**
   * Returns the string encoded for HTML use.<p>
   * > becomes &gt; and < becomes &lt;  
   */
  public static String HTMLEncode(String str) {
    if (null == str) return "null";
    StringBuffer strbuf = new StringBuffer(str.length());
    char tmp;
    for (int i = 0; i < str.length(); i++) {
      tmp = str.charAt(i);
      if ('<' == tmp) {
	strbuf.append("&lt;");
      } else if ('>' == tmp) {
	strbuf.append("&gt;");
      } else {
	strbuf.append(tmp);
      }
    }
    return strbuf.toString();
  }

  /**
   * Returns common prefix for a Vector of strings. This is very useful for
   * determining a commong prefix for a set of paths.
   */
  public static String commonPrefix(Vector v) {
    return commonPrefix(v.elements());
  }

  /**
   * Returns common prefix for an Enumeration of strings.
   */
  public static String commonPrefix(Enumeration en) {
    if (null == en || ! en.hasMoreElements()) return "";
    String common = (String)en.nextElement();
    String str = null;
    int i, len;
    char[] ar1, ar2;

    ar1 = common.toCharArray();
    while (en.hasMoreElements()) {
      str = (String)en.nextElement();
      ar2 = str.toCharArray();
      if (str.startsWith(common)) continue;
      len = common.length();
      if (len > str.length()) len = str.length();
      for (i = 0; i < len; i++) {
	if (ar1[i] != ar2[i]) break;
      }
      if (0 == i) return "";
      common = common.substring(0,i);
      ar1 = common.toCharArray();
    }
    if (-1 != (i = common.indexOf('#'))) common = common.substring(0,i);
    if (-1 != (i = common.indexOf('@'))) common = common.substring(0,i);
    return common;
  }

  /**
   * Returns the change number portion of a depot path, if there is
   * a valid one found. Otherwise, it returns -1.
   */
  public final static int getChangeFromPath(String path) {
    int i = path.indexOf('@');
    if (0 > i) return -1;
    try {
      return Integer.valueOf(path.substring(i+1)).intValue();
    } catch (NumberFormatException ex) {
      return -1;
    }
  }

  /**
   * Cleans up after the package has been used. This stops any running threads
   * and releases any objects for garbage collection.
   */
  public static void cleanUp() {
    HashDecay.stopAll();
    System.gc();
  }

  /**
   * @deprecated  Useful for testing, but should not be documented.
   */
  public static void main(String[] argv) {
    Vector v = new Vector(argv.length);
    for (int i = 0; i < argv.length; i++) {
      v.addElement(argv[i]);
      System.out.println(argv[i]+": "+getChangeFromPath(argv[i]));
    }
    System.out.println("Common: "+commonPrefix(v));
  }
}
